#include <iostream>
#include <vector>
#include <string>
#include <cstdlib>
#include <math.h>
#include <time.h>

using std::cin;
using std::cout;
using std::cerr;
using std::endl;
using std::string;
using std::vector;

enum foot_locations
{
  a_left,
  a_down,
  a_up,
  a_right,
  b_left,
  b_down,
  b_up,
  b_right,
  num_foot_locations
};

enum foot_names
{
  left,
  right,
  num_feet
};

struct player_stance
{
  int foot_locations[2];
  int next_foot;
};

enum constraint_types
{
  move_dist,
  stretch_dist,
  facing_change,
  facing_away,
  body_x_motion,
  body_y_motion,
  body_motion,
  body_x_pos,
  body_y_pos,
  goalness,
  num_constraint_types
};

enum variation_types
{
  no_variation,
  wave_sin,
  wave_cos,
  wave_tan,
  num_variation_types
};

struct variation_data_t
{
  int type;
  double offset;
  double phase;
  double amplitude;
  double frequency;
  // frequency = pi / 180 is one step per degree
  double eval(double step)
  {
    switch(type)
    {
      case no_variation:
        return offset;
      case wave_sin:
        return offset + (amplitude * sin(phase + (frequency * step)));
      case wave_cos:
        return offset + (amplitude * cos(phase + (frequency * step)));
      case wave_tan:
        return offset + (amplitude * tan(phase + (frequency * step)));
    }
    return offset;
  }
};

char const* const constraint_type_names[num_constraint_types]=
{
  "move_dist",
  "stretch_dist",
  "facing_change",
  "facing_away",
  "body_x_motion",
  "body_y_motion",
  "body_motion",
  "body_x_pos",
  "body_y_pos",
  "goalness",
};

struct constraint
{
  int type;
  // TODO:  Switch to using a comparison function the way the STL does.
  bool is_max_constraint;  // A constraint can be a max or a min.
  bool has_edge_case;
  double current_value;
  double edge_sensitivity;  // A case that is betweeen edge_sensitivity and
  // value is considered an edge condition.
  double goal_value;
  double change_rate;
  double after_edge_value;
};

struct weight_changer_t
{
  int type; // Same type values as constraint types.
  bool is_gt_compare;
  double compare_value;
  int weight_change_amount;
  variation_data_t weight_change_variation;
};

struct goal_t
{
  int type; // Same as constraint types.
  double value;
  // TODO:  Switch to using function objects.
  bool (*reach_check)(goal_t* g, player_stance& stance, vector<double>& step_stats);
  void (*reach_response)(goal_t* g, player_stance& stance, vector<double>& step_stats);
};

struct location_candidate_t
{
  int location;
  int weight;
};

struct config_data_t
{
  double bpm;
  size_t notes_per_measure;
  double stream_length_in_minutes;
  double measures;
  int empty_measures;
  bool use_random_seed;
  bool singles_only;
  bool file_header;
  string chart_difficulty;
  string chart_author;
  int chart_foot_rating;
  vector<constraint> constraints;
  vector<weight_changer_t> weight_changers;
};

double const panel_separations[num_foot_locations][num_foot_locations]=
{
  //         aL,    aD,    aU,    aR,    bL,    bD,    bU,    bR
  /* aL */ {0.0,   1.414, 1.414, 2.0,   3.0,   4.123, 4.123, 5.0},
  /* aD */ {1.414, 0.0,   2.0,   1.414, 2.236, 3.0,   3.606, 4.123},
  /* aU */ {1.414, 2.0,   0.0,   1.414, 2.236, 3.606,   3.0, 4.123},
  /* aR */ {2.0,   1.414, 1.414, 0.0,   1.0,   2.236, 2.236, 3.0},
  /* bL */ {3.0,   2.236, 2.236, 1.0,   0.0,   1.414, 1.414, 2.0},
  /* bD */ {4.123, 3.0,   3.606, 2.236, 1.414, 0.0,   2.0,   1.414},
  /* bU */ {4.123, 3.606, 3.0,   2.236, 1.414, 2.0,   0.0,   1.414},
  /* bR */ {5.0,   4.123, 4.123, 3.0,   2.0,   1.414, 1.414, 5.0}
};

// The 0/2pi line will be at the back, to make anything crossing that be thrown out.

double const facing_directions[num_foot_locations][num_foot_locations]=
{
  //         aL,     aD,     aU,     aR,     bL,     bD,     bU,     bR
  /* aL */ {3.1416, 2.3562, 3.9270, 3.1416, 3.1416, 2.8966, 3.3866, 3.1416},
  /* aD */ {5.4978, 3.1416, 4.7124, 3.9270, 3.6052, 3.1416, 3.7296, 3.3866},
  /* aU */ {0.7855, 1.5708, 3.1416, 2.3562, 2.6780, 2.5536, 3.1416, 2.8966},
  /* aR */ {6.2832, 0.7855, 5.4978, 3.1416, 3.1416, 2.6780, 3.6052, 3.1416},
  /* bL */ {6.2832, 0.4637, 5.8195, 6.2832, 3.1416, 2.3562, 3.9270, 3.1416},
  /* bD */ {6.0382, 6.2832, 5.6952, 5.8195, 5.4978, 3.1416, 4.7124, 3.9270},
  /* bU */ {0.2450, 0.5880, 6.2832, 0.4637, 0.7855, 1.5708, 3.1416, 2.3562},
  /* bR */ {6.2832, 0.2450, 6.0382, 6.2832, 6.2832, 0.7855, 5.4980, 3.1416},
};

double const body_positions[num_foot_locations][num_foot_locations][2]=
{
  //           aL,           aD,         aU,          aR,
  //           bL,           bD,         bU,          bR
  /* aL */ {{-2.5, 1.5}, {-2, 1}, {-2, 2}, {-1.5, 1.5},
            {-1, 1.5}, {-0.5, 1}, {-0.5, 2}, {0, 1.5}},
  /* aD */ {{-2, 1}, {-1.5, 0.5}, {-1.5, 1.5}, {-1, 1},
            {-0.5, 1}, {0, 0.5}, {0, 1.5}, {0.5, 1}},
  /* aU */ {{-2, 2}, {-1.5, 1.5}, {-1.5, 2.5}, {-1, 2},
            {-0.5, 2}, {0, 1.5}, {0, 2.5}, {0.5, 2}},
  /* aR */ {{-1.5, 1.5}, {-1, 1}, {-1, 2}, {-0.5, 1.5},
            {0, 1.5}, {0.5, 1}, {0.5, 2}, {1, 1.5}},
  /* bL */ {{-1, 1.5}, {-0.5, 1}, {-0.5, 2}, {0, 1.5},
            {0.5, 1.5}, {1, 1}, {1, 2}, {1.5, 1.5}},
  /* bD */ {{-0.5, 1}, {0, 0.5}, {0, 1.5}, {0.5, 1},
            {1, 1}, {1.5, 0.5}, {1.5, 1.5}, {2, 1}},
  /* bU */ {{-0.5, 2}, {0, 1.5}, {0, 2.5}, {0.5, 2},
            {1, 2}, {1.5, 1.5}, {1.5, 2.5}, {2, 2}},
  /* bR */ {{0, 1.5}, {0.5, 1}, {0.5, 2}, {1, 1.5},
            {1.5, 1.5}, {2, 1}, {2, 2}, {2.5, 1.5}}
};

bool back_and_forth_goal_check(goal_t* g, player_stance& stance, vector<double>& step_stats)
  {
    if(g->value < 0.0)
    {
      return step_stats[body_x_pos] < g->value;
    }
    else
    {
      return step_stats[body_x_pos] > g->value;
    }
  }

void back_and_forth_goal_response(goal_t* g, player_stance& stance, vector<double>& step_stats)
  {
    g->value*= -1.0;
  }

void calc_step_stats(player_stance const& cur_stance, int next_location, goal_t& goal, vector<double>& ret)
  {
    // The stats will be pushed onto the vector in the same order as they
    // appear in constraint_types.
    int cur_foot_location= cur_stance.foot_locations[cur_stance.next_foot];
    double move_distance= panel_separations[cur_foot_location][next_location];
    ret.push_back(move_distance);
    int afoot_loc= cur_stance.foot_locations[!cur_stance.next_foot];
    int bfoot_loc= next_location;
    double stretch_distance= panel_separations[afoot_loc][bfoot_loc];
    ret.push_back(stretch_distance);
    double cur_facing= facing_directions[cur_stance.foot_locations[left]]
      [cur_stance.foot_locations[right]];
    double next_facing= cur_facing;
    switch(cur_stance.next_foot)
    {
      case left:
        next_facing= facing_directions[next_location]
          [cur_stance.foot_locations[right]];
        break;
      case right:
        next_facing= facing_directions[cur_stance.foot_locations[left]]
          [next_location];
        break;
      default:
        break;
    }
    double facing_change= fabs(next_facing - cur_facing);
    ret.push_back(facing_change);
    double facing_away= fabs(next_facing - M_PI);
    ret.push_back(facing_away);
    double old_x= body_positions[cur_stance.foot_locations[0]]
      [cur_stance.foot_locations[1]][0];
    double old_y= body_positions[cur_stance.foot_locations[0]]
      [cur_stance.foot_locations[1]][1];
    double new_x= body_positions[afoot_loc][bfoot_loc][0];
    double new_y= body_positions[afoot_loc][bfoot_loc][1];
    double x_diff= new_x - old_x;
    double y_diff= new_y - old_y;
    ret.push_back(fabs(x_diff));
    ret.push_back(fabs(y_diff));
    double body_motion_dist= sqrt((x_diff * x_diff) + (y_diff * y_diff));
    ret.push_back(body_motion_dist);
    ret.push_back(new_x);
    ret.push_back(new_y);
    if(goal.type >= 0 && goal.type < goalness)
    {
      double goalness= ret[goal.type] / goal.value;
      ret.push_back(goalness);
    }
  }

bool is_position_allowed(player_stance const& cur_stance,
  location_candidate_t& cand, vector<constraint>& constraints,
  vector<weight_changer_t>& weights, goal_t& goal, bool explain_reject, bool singles_only,
  size_t step_count)
  {
    if(singles_only)
    {
      if(cand.location >= b_left)
      {
        return false;
      }
    }
    // disallow footswitches
    if(cand.location == cur_stance.foot_locations[!cur_stance.next_foot])
    {
      if(explain_reject)
      {
        cerr << cand.location << "Other foot is already in that position." << endl;
      }
      return false;
    }
    vector<double> step_stats;
    calc_step_stats(cur_stance, cand.location, goal, step_stats);
    for(vector<constraint>::iterator c= constraints.begin();
        c != constraints.end(); ++c)
    {
      if(c->type >= 0 && c->type < num_constraint_types)
      {
        if(c->is_max_constraint)
        {
          if(step_stats[c->type] > c->current_value)
          {
            if(explain_reject)
            {
              cerr << cand.location << constraint_type_names[c->type] << ": "
                << step_stats[move_dist] << " > " << c->current_value << endl;
            }
            return false;
          }
        }
        else
        {
          if(step_stats[c->type] < c->current_value)
          {
            if(explain_reject)
            {
              cerr << cand.location << constraint_type_names[c->type] << ": "
                << step_stats[move_dist] << " < " << c->current_value << endl;
            }
            return false;
          }
        }
      }
    }
    for(vector<weight_changer_t>::iterator w= weights.begin();
        w != weights.end(); ++w)
    {
      if(w->type >= 0 && w->type < num_constraint_types)
      {
        double real_weight_change= w->weight_change_amount +
          w->weight_change_variation.eval(step_count);
        if(explain_reject)
        {
          cerr << cand.location << " rwc: " << w->weight_change_amount
            << " + " << w->weight_change_variation.eval(step_count) << " = "
            << real_weight_change << endl;
        }
        
        if(w->is_gt_compare)
        {
          if(step_stats[w->type] > w->compare_value)
          {
            cand.weight+= real_weight_change;
          }
        }
        else
        {
          if(step_stats[w->type] < w->compare_value)
          {
            cand.weight+= real_weight_change;
          }
        }
      }
    }
    return true;
  }

void gen_stream(vector<size_t>& ret, config_data_t& config)
  {
    size_t length= static_cast<size_t>(config.notes_per_measure * config.measures);
    cerr << "Generting stream of " << length << " steps." << endl;
    player_stance cur_stance;
    vector<constraint>& constraints= config.constraints;
    vector<weight_changer_t>& weight_changers= config.weight_changers;
    goal_t goal;
    goal.type= body_x_pos;
    goal.value= 1.5;
    goal.reach_check= &back_and_forth_goal_check;
    goal.reach_response= &back_and_forth_goal_response;

    if(config.singles_only)
    {
      cur_stance.foot_locations[left]= a_left;
      cur_stance.foot_locations[right]= a_right;
     }
    else
    {
      cur_stance.foot_locations[left]= a_right;
      cur_stance.foot_locations[right]= b_left;
    }
    cur_stance.next_foot= rand() & 1;
    cerr << "Starting with foot " << cur_stance.next_foot << endl;
    // Make sure it starts on the middle two arrows.
    ret.push_back(cur_stance.foot_locations[cur_stance.next_foot]);
    ret.push_back(cur_stance.foot_locations[!cur_stance.next_foot]);
    cerr << "Tolerances set, beginning placement." << endl;
    while(ret.size() < length)
    {
      vector<location_candidate_t> allowed_next_locations;
      int total_weight= 0;
      for(int l= 0; l < num_foot_locations; ++l)
      {
        location_candidate_t cand;
        cand.location= l;
        cand.weight= 10;
        if(is_position_allowed(cur_stance, cand, constraints, weight_changers,
            goal, false, config.singles_only, ret.size()))
        {
          allowed_next_locations.push_back(cand);
          total_weight+= cand.weight;
        }
      }
      int next_loc= cur_stance.foot_locations[cur_stance.next_foot];
      if(allowed_next_locations.size() == 0)
      {
        cerr << "Stream params too tight, cannot continue stream." << endl;
        cerr << "Position:  " << cur_stance.foot_locations[0]+1 << ", "
          << cur_stance.foot_locations[1]+1 << "  Next foot: "
          << cur_stance.next_foot << endl;
        for(int l= 0; l < num_foot_locations; ++l)
        {
          location_candidate_t cand;
          cand.location= l;
          cand.weight= 10;
          if(is_position_allowed(cur_stance, cand, constraints,
              weight_changers, goal, true, config.singles_only, ret.size()))
          {
            cerr << "location " << l
              << " not rejected when explaining.  Investigate." << endl;
          }
        }
        return;
      }
      else if(allowed_next_locations.size() == 1)
      {
        next_loc= allowed_next_locations[0].location;
      }
      else
      {
        if(total_weight <= 0)
        {
          cerr << "Total weight too small to use." << endl;
          for(int l= 0; l < num_foot_locations; ++l)
          {
            location_candidate_t cand;
            cand.location= l;
            cand.weight= 10;
            if(is_position_allowed(cur_stance, cand, constraints,
                weight_changers, goal, true, config.singles_only, ret.size()))
            {
              cerr << "location " << l
                << " not rejected when explaining.  Investigate." << endl;
            }
          }
          return;
          int picked_loc= rand() % allowed_next_locations.size();
          next_loc= allowed_next_locations[picked_loc].location;
        }
        else
        {
          int picked_weight= rand() % total_weight;
          int weight_so_far= 0;
          for(vector<location_candidate_t>::iterator cand=
                allowed_next_locations.begin();
              cand != allowed_next_locations.end(); ++cand)
          {
            weight_so_far+= cand->weight;
            if(weight_so_far > picked_weight)
            {
              next_loc= cand->location;
              break;
            }
          }
        }
      }
      ret.push_back(next_loc);
      vector<double> step_stats;
      calc_step_stats(cur_stance, next_loc, goal, step_stats);
      for(vector<constraint>::iterator c= constraints.begin();
          c != constraints.end(); ++c)
      {
        if(c->has_edge_case && c->type >= 0 && c->type < num_constraint_types)
        {
          bool is_edge= false;
          if(c->is_max_constraint)
          {
            if(step_stats[c->type] > c->edge_sensitivity)
            {
              is_edge= true;
            }
          }
          else
          {
            if(step_stats[c->type] < c->edge_sensitivity)
            {
              is_edge= true;
            }
          }
          if(is_edge)
          {
            c->current_value= c->after_edge_value;
          }
          else
          {
            c->current_value+= c->change_rate;
            if((c->is_max_constraint && c->current_value > c->goal_value) ||
              (!c->is_max_constraint && c->current_value < c->goal_value))
            {
              c->current_value= c->goal_value;
            }
          }
        }
      }
      cur_stance.foot_locations[cur_stance.next_foot]= next_loc;
      cur_stance.next_foot= !cur_stance.next_foot;
    }
  }

bool input_double(string const& name, double& d)
  {
    double f;
    if(!(cin >> f))
    {
      cerr << "Value for " << name << " must be a number." << endl;
      return false;
    }
    d= f;
    return true;
  }

bool input_int(string const& name, int& d)
  {
    int f;
    if(!(cin >> f))
    {
      cerr << "Value for " << name << " must be an integer." << endl;
      return false;
    }
    d= f;
    return true;
  }

bool input_size(string const& name, size_t& d)
  {
    int f;
    if(!(cin >> f))
    {
      cerr << "Value for " << name << " must be a positive integer." << endl;
      return false;
    }
    d= f;
    return true;
  }

bool input_bool(string const& name, bool& d)
  {
    string s;
    cin >> s;
    if(s == "true" || s == "1")
    {
      d= true;
    }
    else if(s == "false" || s == "0")
    {
      d= false;
    }
    else
    {
      cerr << "Value for " << name << " must be a boolean (true, false, 0, or 1)." << endl;
      return false;
    }
    return true;
  }

#define RETURN_IF_FAIL(condition) if(!(condition)) { return false; };

bool parse_config_file_input(config_data_t& config)
  {
    // A simple keyword based parser.
    string next_type;
    while(cin >> next_type)
    {
      if(next_type == "bpm")
      {
        RETURN_IF_FAIL((input_double("bpm", config.bpm)));
      }
      else if(next_type == "npm")
      {
        RETURN_IF_FAIL((input_size("npm", config.notes_per_measure)));
      }
      else if(next_type == "minutes")
      {
        RETURN_IF_FAIL((input_double("minutes", config.stream_length_in_minutes)));
      }
      else if(next_type == "empty")
      {
        RETURN_IF_FAIL((input_int("empty measures", config.empty_measures)));
      }
      else if(next_type == "use_random_seed")
      {
        config.use_random_seed= true;
      }
      else if(next_type == "dont_use_random_seed")
      {
        config.use_random_seed= false;
      }
      else if(next_type == "singles")
      {
        config.singles_only= true;
      }
      else if(next_type == "doubles")
      {
        config.singles_only= false;
      }
      else if(next_type == "fake_header")
      {
        config.file_header= true;
      }
      else if(next_type == "no_header")
      {
        config.file_header= false;
      }
      else if(next_type == "diff")
      {
        cin >> config.chart_difficulty;
      }
      else if(next_type == "author")
      {
        config.chart_author.clear();
        string nnp;
        cin >> nnp;
        while(nnp[nnp.size()-1] == '\\')
        {
          nnp[nnp.size()-1]= ' ';
          config.chart_author+= nnp;
          cin >> nnp;
          if(!cin)
          {
            cerr << "Reading config failed while trying to read extended name." << endl;
            return false;
          }
        }
        config.chart_author+= nnp;
      }
      else if(next_type == "feet")
      {
        RETURN_IF_FAIL((input_int("feet", config.chart_foot_rating)));
      }
      else if(next_type == "constraint")
      {
        constraint c;
        cin >> next_type;
        c.type= -1;
        for(int n= 0; n < num_constraint_types; ++n)
        {
          if(next_type == constraint_type_names[n])
          {
            c.type= n;
            break;
          }
        }
        if(c.type == -1)
        {
          cerr << next_type << " is not the name of a constraint type." << endl;
          return false;
        }
        c.has_edge_case= false;
        while(cin >> next_type)
        {
          if(next_type == "max")
          {
            c.is_max_constraint= true;
          }
          else if(next_type == "min")
          {
            c.is_max_constraint= false;
          }
          else if(next_type == "cur")
          {
            RETURN_IF_FAIL((input_double("cur", c.current_value)));
          }
          else if(next_type == "edge")
          {
            RETURN_IF_FAIL((input_double("edge", c.edge_sensitivity)));
          }
          else if(next_type == "goal")
          {
            RETURN_IF_FAIL((input_double("constraint goal", c.goal_value)));
            c.has_edge_case= true;
          }
          else if(next_type == "rate")
          {
            RETURN_IF_FAIL((input_double("rate", c.change_rate)));
          }
          else if(next_type == "after_edge")
          {
            RETURN_IF_FAIL((input_double("after_edge", c.after_edge_value)));
          }
          else if(next_type == "end")
          {
            break;
          }
        }
        if(c.type == facing_change || c.type == facing_away)
        {
          c.current_value*= M_PI;
          c.edge_sensitivity*= M_PI;
          c.goal_value*= M_PI;
          c.change_rate*= M_PI;
          c.after_edge_value*= M_PI;
        }
        config.constraints.push_back(c);
      }
      else if(next_type == "weight")
      {
        weight_changer_t w;
        cin >> next_type;
        w.type= -1;
        for(int n= 0; n < num_constraint_types; ++n)
        {
          if(next_type == constraint_type_names[n])
          {
            w.type= n;
            break;
          }
        }
        if(w.type == -1)
        {
          cerr << next_type << " is not the name of a constraint type." << endl;
          return false;
        }
        while(cin >> next_type)
        {
          if(next_type == ">")
          {
            w.is_gt_compare= true;
          }
          else if(next_type == "<")
          {
            w.is_gt_compare= false;
          }
          else if(next_type == "value")
          {
            RETURN_IF_FAIL((input_double("weight value", w.compare_value)));
          }
          else if(next_type == "change")
          {
            RETURN_IF_FAIL((input_int("weight change", w.weight_change_amount)));
          }
          else if(next_type == "varia")
          {
            cin >> next_type;
            if(next_type == "sin")
            {
              w.weight_change_variation.type= wave_sin;
            }
            else if(next_type == "cos")
            {
              w.weight_change_variation.type= wave_cos;
            }
            else if(next_type == "tan")
            {
              w.weight_change_variation.type= wave_tan;
            }
            RETURN_IF_FAIL((input_double("varia offset", w.weight_change_variation.offset)));
            RETURN_IF_FAIL((input_double("varia phase", w.weight_change_variation.phase)));
            RETURN_IF_FAIL((input_double("varia amplitude", w.weight_change_variation.amplitude)));
            RETURN_IF_FAIL((input_double("varia frequency", w.weight_change_variation.frequency)));
          }
          else if(next_type == "noveria")
          {
            w.weight_change_variation.type= no_variation;
          }
          else if(next_type == "end")
          {
            break;
          }
        }
        config.weight_changers.push_back(w);
      }
    }
    return true;
  }

int main()
  {
    vector<size_t> stream;
    config_data_t config;
    config.file_header= true;
    config.chart_difficulty= "Challenge";
    config.chart_author= "Stepmania Science";
    config.chart_foot_rating= 13;
    config.empty_measures= 0;
    if(!parse_config_file_input(config))
    {
      cerr << "Bad config file, aborting." << endl;
      return 0;
    }
    if(config.use_random_seed)
    {
      srand(time(NULL));
    }
    else
    {
      srand(0);
    }
    config.measures= config.stream_length_in_minutes * (config.bpm / 4.0);
    cerr << config.bpm << " bpm, " << config.notes_per_measure << " npm, "
      << config.stream_length_in_minutes << " minutes, " << config.measures
      << " measures." << endl;
    cerr << "Generating stream." << endl;
    clock_t start_time= clock();
    gen_stream(stream, config);
    clock_t end_time= clock();
    double time_used= static_cast<double>(end_time - start_time) / CLOCKS_PER_SEC;
    cerr << "Stream generated in " << time_used << endl;
    if(config.file_header)
    {
      cout << "#TITLE:Kurumi Ponchio;" << endl
        << "#SUBTITLE:Len Cover;" << endl
        << "#ARTIST:ギガP feat. 鏡音レン;" << endl
        << "#TITLETRANSLIT:Kurumi Ponchio;" << endl
        << "#SUBTITLETRANSLIT:;" << endl
        << "#ARTISTTRANSLIT:Kagamine Len;" << endl
        << "#GENRE:;" << endl
        << "#CREDIT:StarCharm;" << endl
        << "#BANNER:;" << endl
        << "#BACKGROUND:;" << endl
        << "#LYRICSPATH:;" << endl
        << "#CDTITLE:;" << endl
        << "#MUSIC:Kurumi Ponchio Kagamine Len.ogg;" << endl
        << "#OFFSET:-0.140;" << endl
        << "#SAMPLESTART:110.640;" << endl
        << "#SAMPLELENGTH:12.000;" << endl
        << "#SELECTABLE:YES;" << endl
        << "#DISPLAYBPM:140.000;" << endl
        << "#BPMS:0.000=140.000;" << endl
        << "#STOPS:;" << endl
        << "HANGES:;" << endl
        << "#KEYSOUNDS:;" << endl
        << "#ATTACKS:;" << endl
        << "" << endl;
    }
    if(config.singles_only)
    {
      cout << "//---------------dance-single - " << config.chart_author
        << "----------------" << endl;
    }
    else
    {
      cout << "//---------------dance-double - " << config.chart_author
        << "----------------" << endl;
    }
    cout << "#NOTES:" << endl;
    if(config.singles_only)
    {
      cout << "     dance-single:" << endl;
    }
    else
    {
      cout << "     dance-double:" << endl;
    }
    cout << "     " << config.chart_author << ":" << endl
      << "     " << config.chart_difficulty << ":" << endl
      << "     " << config.chart_foot_rating << ":" << endl
      << "     0.853,1.000,0.008,0.591,0.894,1291.000,2.000,142.000,99.000,32.000,16.000,0.853,1.000,0.008,0.591,0.894,1291.000,2.000,142.000,99.000,32.000,16.000:" << endl;
    for(int n= 0; n < config.empty_measures; ++n)
    {
      cout << "0000" << endl
        << "0000" << endl
        << "0000" << endl
        << "0000" << endl
        << "," << endl;
    }

    for(size_t n= 0; n < stream.size(); ++n)
    {
      for(size_t l= 0; l < num_foot_locations; ++l)
      {
        if(config.singles_only && l >= b_left)
        {
          break;
        }
        if(stream[n] == l)
        {
          cout << 1;
        }
        else
        {
          cout << 0;
        }
      }
      cout << endl;
      if(n % config.notes_per_measure == config.notes_per_measure - 1)
      {
        if(n == stream.size() - 1)
        {
          cout << ";" << endl;
        }
        else
        {
          cout << "," << endl;
        }
      }
    }
  }
